-- run as 'lua xtable-benchmark.lua 100000' etc.

core = require "sort_core"
p2sort = dofile "p2sort.lua"
ssort = dofile "stablesort.lua"
worstcase = require"worstcase"

--local ceil=math.ceil
--chunk_size = function(n)
--   if n<=16 then return n end
--   repeat n = ceil(n/2) until n<=16
--   return n
--end

max_chunk_size=tonumber(arg[2]) or 8
p2sortn = dofile "p2sort.lua"
ssortn= dofile "stablesort.lua"

X=require"xtable"
B=X.block
T=X.tuple

function sorted(a,cmp,l,r)
   l, r = l or 1, r or #a
   if cmp then for k=l,r-1 do if cmp(a[k+1],a[k]) then return false end end
   else for k=l,r-1 do if a[k+1]<a[k] then return false end end end
   return true
end

function cmp1(a,b) return a[1]<b[1] end
function cmp2(a,b) return a[2]<b[2] end

function sorttest(a,cmp1,cmp2)
--- sorts a by cmp1 using three methods.
-- if cmp2 is given, tests whether a is sorted by cmp2 and whether
-- this ordering is preserved for equals under cmp1   
   local sorted2 = cmp2 and sorted(a,cmp2)
   local cmp = function(x,y)
      return cmp1(x,y) or not cmp1(y,x) and cmp2(x,y)
   end
   local b=setmetatable({},{__index=B,__tostring=B.concat})
   local function test(method,name)
      for k=1,#a do b[k]=a[k] end
      tic=os.clock(); method(b,cmp1); tic=os.clock()-tic
      if not sorted(b,cmp1) or nil then tic=0/0
      elseif sorted2 and not sorted(b,cmp) then tic=-tic
      end
      return tic
   end
   ------
   for method=1,#methods do 
      result[method][testfunc]=test(methods[method][1]) 
   end
end

local ceil, random, sqrt, abs = math.ceil, math.random, math.sqrt, math.abs

nmax = arg[1] or 10
-- nmax = 20 * math.ceil(nmax/20)
swaps = math.ceil(math.sqrt(nmax))

M={__tostring = function(t) return '('..table.concat(t,':')..')' end}
a=setmetatable({},{__index=B,__tostring=B.concat})

methods = { 
   {table.sort,"table.sort"},
--   {p2sort,"power-of-2 merge sort"},
--   {p2sortn,"ditto, with chunk size "..max_chunk_size},
   {ssort,"stablesort"},
--   {ssortn,"ditto, with chunk size "..max_chunk_size}
}

result = {}
for method=1,#methods do result[method]={} end

testfunc=1
print (("== test"..testfunc..": %d random duples"):format(nmax))
for k=1,nmax do a[k]=setmetatable({random(),random()},M) end
sorttest(a,cmp2,cmp1)

testfunc=testfunc+1
print (("== test"..testfunc..": %d cut deck"):format(nmax))
for k=1,nmax do a[k]=(k+(nmax-1)/2)%nmax+1 end
sorttest(a)

testfunc=testfunc+1
print (("== test"..testfunc..": %d sorted elements with %d random swaps"):
  format(nmax,swaps))
for k=1,nmax do a[k]=k end
for k=1,swaps do 
   local i,j = random(nmax), random(nmax)
   a[i],a[j] = a[j],a[i]
end
sorttest(a)

testfunc=testfunc+1
print (("== test"..testfunc..": %d duples with 20 distinct values"):format(nmax))
for k=1,nmax do a[k]=setmetatable({k,(59*k)%20},M) end
sorttest(a,cmp2,cmp1)

testfunc=testfunc+1
print (("== test"..testfunc..
   ": %d duples with approximately 20 equals of every value"):format(nmax))
for k=1,nmax do a[k]=setmetatable({k,(59*k)%(nmax/20)},M) end
sorttest(a,cmp2,cmp1)

testfunc=testfunc+1
a=worstcase(15); nmax=#a;
print (("== test"..testfunc..
   ": permutation of 1:%d designed as a worst case for quicksort"):format(nmax))
sorttest(a)

print("\n                   Time in seconds, negative time means sort was unstable\n")

io.write(("%29s"):format"")
for k=1,#result[1] do io.write(("     #%1i"):format(k)) end
io.write("    Total\n")
for method=1,#methods do 
   io.write (("%30s"):format(methods[method][2]))
   local res = result[method]
   local total = 0
   for k=1,#res do 
      io.write(("%7.2f"):format(res[k])) 
      total = total+abs(res[k])
   end
   io.write(("%8.2f"):format(total),"\n")
end

